
#include "../h/param.h"
#include "../h/systm.h"
#include "../h/errno.h"
#include "../h/dir.h"
#include "../h/buf.h"
#include "../h/conf.h"
#include "../h/proc.h"

/*
 * The following several routines allocate and free
 * buffers with various side effects.  In general the
 * arguments to an allocate routine are a device and
 * a block number, and the value is a pointer to
 * to the buffer header; the buffer is marked "busy"
 * so that no one else can touch it.  If the block was
 * already in core, no I/O need be done; if it is
 * already busy, the process waits until it becomes free.
 * The following routines allocate a buffer:
 *	getblk
 *	bread
 *	breada
 *	baddr	(if it is incore)
 * Eventually the buffer must be released, possibly with the
 * side effect of writing it out, by using one of
 *	bwrite
 *	bdwrite
 *	bawrite
 *	brelse
 */

/*
 * Read in (if necessary) the block and return a buffer pointer.
 */
struct buf *
bread(dev, blkno)
	dev_t dev;
	daddr_t blkno;
{
	register struct buf *bp = getblk(dev, blkno);

	if ((bp->b_flags&B_DONE) == 0) {
		bp->b_flags |= B_READ;
		bp->b_bcount = BSIZE;
		(*bdevsw[major(dev)].d_strategy)(bp);
		iowait(bp);
	}
	return(bp);
}

/*
 * Read in the block, like bread, but also start I/O on the
 * read-ahead block (which is not allocated to the caller)
 */
struct buf *
breada(dev, blkno, rablkno)
	dev_t dev;
	daddr_t blkno, rablkno;
{
	register struct buf *bp, *rabp;

	bp = NULL;
	if (!incore(dev, blkno)) {
		bp = getblk(dev, blkno);
		if ((bp->b_flags&B_DONE) == 0) {
			bp->b_flags |= B_READ;
			bp->b_bcount = BSIZE;
			(*bdevsw[major(dev)].d_strategy)(bp);
		}
	}
	if (rablkno && !incore(dev, rablkno)) {
		rabp = getblk(dev, rablkno);
		if (rabp->b_flags & B_DONE)
			brelse(rabp);
		else {
			rabp->b_flags |= B_READ|B_ASYNC;
			rabp->b_bcount = BSIZE;
			(*bdevsw[major(dev)].d_strategy)(rabp);
		}
	}
	if (bp == NULL)
		return (bread(dev, blkno));

	iowait(bp);
	return (bp);
}

/*
 * Write the buffer, waiting for completion.
 * Then release the buffer.
 */
bwrite(bp)
	register struct buf *bp;
{
	register flag;

	flag = bp->b_flags;
	bp->b_flags &= ~(B_READ | B_DONE | B_ERROR | B_DELWRI | B_AGE);
	bp->b_bcount = BSIZE;
	(*bdevsw[major(bp->b_dev)].d_strategy)(bp);

	if ((flag&B_ASYNC) == 0) {
		iowait(bp);
		brelse(bp);

	} else if (flag & B_DELWRI) {
		bp->b_flags |= B_AGE;

	} else if (bp->b_flags&B_ERROR)
		errno = EIO;
}

/*
 * Release the buffer, marking it so that if it is grabbed
 * for another purpose it will be written out before being
 * given up (e.g. when writing a partial block where it is
 * assumed that another write for the same block will soon follow).
 * This can't be done for magtape, since writes must be done
 * in the same order as requested.
 */
bdwrite(bp)
	register struct buf *bp;
{
	register struct buf *dp;

	dp = bdevsw[major(bp->b_dev)].d_tab;
	if (dp->b_flags & B_TAPE)
		bawrite(bp);
	else {
		bp->b_flags |= B_DELWRI | B_DONE;
		brelse(bp);
	}
}

/*
 * Release the buffer, start I/O on it, but don't wait for completion.
 */
bawrite(bp)
	register struct buf *bp;
{

	bp->b_flags |= B_ASYNC;
	bwrite(bp);
}

/*
 * release the buffer, with no I/O implied.
 */
brelse(bp)
	register struct buf *bp;
{
	register struct buf **backp;
	register s;

	if (bp->b_flags&B_WANTED)
		wakeup(bp);

	if (bfreelist.b_flags&B_WANTED) {
		bfreelist.b_flags &= ~B_WANTED;
		wakeup(&bfreelist);
	}
	if (bp->b_flags&B_ERROR)
		bp->b_dev = NODEV;  /* no assoc. on error */

	s = splbio();
	if (bp->b_flags & B_AGE) {
		backp = &bfreelist.av_forw;
		(*backp)->av_back = bp;
		bp->av_forw = *backp;
		*backp = bp;
		bp->av_back = &bfreelist;
	} else {
		backp = &bfreelist.av_back;
		(*backp)->av_forw = bp;
		bp->av_back = *backp;
		*backp = bp;
		bp->av_forw = &bfreelist;
	}
	bp->b_flags &= ~(B_WANTED|B_BUSY|B_ASYNC|B_AGE);
	splx(s);
}

/*
 * See if the block is associated with some buffer
 * (mainly to avoid getting hung up on a wait in breada)
 */
incore(dev, blkno)
	dev_t dev;
	daddr_t blkno;
{
	register struct buf *bp;
	register struct buf *dp;

	dp = bdevsw[major(dev)].d_tab;
	for (bp = dp->b_forw; bp != dp; bp = bp->b_forw)
		if (bp->b_blkno == blkno && bp->b_dev == dev)
			return (1);

	return (0);
}

struct buf *
baddr(dev, blkno)
	dev_t dev;
	daddr_t blkno;
{

	if (incore(dev, blkno))
		return (bread(dev, blkno));

	return (0);
}

/*
 * Assign a buffer for the given block.  If the appropriate
 * block is already associated, return it; otherwise search
 * for the oldest non-busy buffer and reassign it.
 */
struct buf *
getblk(dev, blkno)
	dev_t dev;
	daddr_t blkno;
{
	register struct buf *bp, *dp;

	if (major(dev) >= nblkdev)
		panic("blkdev");

loop:
	spl0();
	dp = bdevsw[major(dev)].d_tab;
	if (dp == NULL)
		panic("devtab");

	for (bp = dp->b_forw; bp != dp; bp = bp->b_forw) {
		if (bp->b_blkno != blkno || bp->b_dev != dev)
			continue;
		splbio();
		if (bp->b_flags&B_BUSY) {
			bp->b_flags |= B_WANTED;
			sleep(bp, PRIBIO+1);
			goto loop;
		}
		spl0();
		notavail(bp);
		return (bp);
	}
	splbio();
	if (bfreelist.av_forw == &bfreelist) {
		bfreelist.b_flags |= B_WANTED;
		sleep(&bfreelist, PRIBIO+1);
		goto loop;
	}
	spl0();

	bp = bfreelist.av_forw;
	notavail(bp);
	if (bp->b_flags & B_DELWRI) {
		bp->b_flags |= B_ASYNC;
		bwrite(bp);
		goto loop;
	}
	bp->b_flags = B_BUSY;
	bp->b_back->b_forw = bp->b_forw;
	bp->b_forw->b_back = bp->b_back;
	bp->b_forw = dp->b_forw;
	bp->b_back = dp;
	dp->b_forw->b_back = bp;
	dp->b_forw = bp;

	bp->b_dev = dev;
	bp->b_blkno = blkno;

	return (bp);
}

/*
 * get an empty block,
 * not assigned to any particular device
 */
struct buf *
geteblk()
{
	register struct buf *bp, *dp;

loop:
	splbio();
	while (bfreelist.av_forw == &bfreelist) {
		bfreelist.b_flags |= B_WANTED;
		sleep(&bfreelist, PRIBIO+1);
	}
	spl0();

	bp = bfreelist.av_forw;
	notavail(bp);
	if (bp->b_flags & B_DELWRI) {
		bp->b_flags |= B_ASYNC;
		bwrite(bp);
		goto loop;
	}
	dp = &bfreelist;

	bp->b_flags = B_BUSY;
	bp->b_back->b_forw = bp->b_forw;
	bp->b_forw->b_back = bp->b_back;
	bp->b_forw = dp->b_forw;
	bp->b_back = dp;
	dp->b_forw->b_back = bp;
	dp->b_forw = bp;

	bp->b_dev = NODEV;

	return (bp);
}

/*
 * Wait for I/O completion on the buffer; return errors
 * to the user.
 */
iowait(bp)
	register struct buf *bp;
{
	splbio();
	while ((bp->b_flags&B_DONE) == 0)
		sleep(bp, PRIBIO);

	spl0();
	if (bp->b_flags&B_ERROR)
		errno = EIO;
}

/*
 * Unlink a buffer from the available list and mark it busy.
 * (internal interface)
 */
notavail(bp)
	register struct buf *bp;
{
	register s;

	s = splbio();
	bp->av_back->av_forw = bp->av_forw;
	bp->av_forw->av_back = bp->av_back;
	bp->b_flags |= B_BUSY;
	splx(s);
}

/*
 * Mark I/O complete on a buffer. If the header
 * indicates a dirty page push completion, the
 * header is inserted into the ``cleaned'' list
 * to be processed by the pageout daemon. Otherwise
 * release it if I/O is asynchronous, and wake 
 * up anyone waiting for it.
 */
iodone(bp)
	register struct buf *bp;
{
	register int s;

	bp->b_flags |= B_DONE;
	if (bp->b_flags&B_ASYNC)
		brelse(bp);
	else {
		bp->b_flags &= ~B_WANTED;
		wakeup(bp);
	}
}

/*
 * Zero the core associated with a buffer.
 */
clrbuf(bp)
	struct buf *bp;
{
	bp->b_resid = 0;
	bzero(bp->b_un.b_addr, BSIZE);
}

/*
 * make sure all write-behind blocks
 * on dev (or NODEV for all)
 * are flushed out.
 * (from umount and update)
 */
bflush(dev)
	dev_t dev;
{
	register struct buf *bp;

loop:
	splbio();
	for (bp = bfreelist.av_forw; bp != &bfreelist; bp = bp->av_forw) {
		if ((bp->b_flags&B_DELWRI)
		 && (dev == NODEV || dev == bp->b_dev)) {
			bp->b_flags |= B_ASYNC;
			notavail(bp);
			bwrite(bp);
			goto loop;
		}
	}
	spl0();
}

/*
 * Raw I/O. The arguments are
 *	The strategy routine for the device
 *	A buffer, which will always be a special buffer
 *	  header owned exclusively by the device for this purpose
 *	The device number
 *	Read/write flag
 */
physio(strat, bp, dev, rw, minphys)
	void (*strat)(); 
	struct buf *bp;
	void (*minphys)();
{
	register int cc;
	char *a;

	splbio();
	while (bp->b_flags&B_BUSY) {
		bp->b_flags |= B_WANTED;
		sleep(bp, PRIBIO+1);
	}
	bp->b_flags = B_BUSY | B_PHYS | rw;
	spl0();

	bp->b_dev = dev;
	bp->b_proc = curproc;
	curproc->p_flag |= SKEEP;

	while (u.u_count) {
		bp->b_flags &= ~B_DONE;
		bp->b_bcount    = u.u_count;
		bp->b_un.b_addr = u.u_base;
		bp->b_blkno     = u.u_offset >> BSHIFT;

		(*minphys)(bp);
		cc = bp->b_bcount;
		(*strat)(bp);

		splbio();
		while ((bp->b_flags & B_DONE) == 0)
			sleep(bp, PRIBIO);
		spl0();

		if (bp->b_flags&B_ERROR) {
			errno = EIO;
			break;
		}
		u.u_base   += cc;
		u.u_offset += cc;
		u.u_count  -= cc;
	}
	curproc->p_flag &= ~SKEEP;
	u.u_count = bp->b_resid;

	bp->b_flags &= ~B_BUSY;
	if (bp->b_flags&B_WANTED) {
		bp->b_flags &= ~B_WANTED;
		wakeup(bp);
	}
}

void
minphys(bp)
	register struct buf *bp;
{
	if (bp->b_bcount > 32 * 1024)
		bp->b_bcount = 32 * 1024;
}

