
#include "../h/param.h"
#include "../h/systm.h"
#include "../h/errno.h"
#include "../h/dir.h"
#include "../h/proc.h"
#include "../h/inode.h"
#include "../h/ino.h"
#include "../h/filsys.h"
#include "../h/mount.h"
#include "../h/conf.h"
#include "../h/buf.h"

/*
 * Look up an inode by device,inumber.
 * If it is in core (in the inode structure),
 * honor the locking protocol.
 * If it is not in core, read it in from the
 * specified device.
 * If the inode is mounted on, perform
 * the indicated indirection.
 * In all cases, a pointer to a locked
 * inode structure is returned.
 *
 * printf warning: no inodes -- if the inode structure is full
 * panic: no imt -- if the mounted file
 *	system is not in the mount table.
 *	"cannot happen"
 */
struct inode *
iget(mp, ino)
	struct mount *mp;
	ino_t ino;
{
	struct inode *ip, *oip;
	struct dinode *dp;
	struct buf *bp;

rescan:
	oip = NULL;
	for (ip = inode; ip < &inode[NINODE]; ip++) {
		if (ip->i_number == ino && ip->i_mp == mp) {
			if (ip->i_flag&ILOCK) {
				ip->i_flag |= IWANT;
				sleep(ip, PINOD);
				goto rescan;
			}
			if (ip->i_flag&IMOUNT) {
				for (mp = mount; mp < &mount[NMOUNT]; ++mp)
					if (mp->m_mpoint == ip) {
						ino = ROOTINO;
						goto rescan;       /* mp changed */
					}
				panic("no imt");
			}
			ip->i_flag |= ILOCK;
			ip->i_count++;
			return (ip);
		}
		if (ip->i_count == 0 && oip == NULL)
			oip = ip;
	}
	if ((ip = oip) == NULL) {
		printf("Inode table overflow\n");
		errno = ENFILE;
		return (NULL);
	}
	ip->i_mp = mp;
	ip->i_number = ino;
	ip->i_flag = ILOCK;
	ip->i_count = 1;
	ip->i_lastr = 0;

	bp = bread(mp->m_dev, itod(ino));
	/*
	 * Check I/O errors
	 */
	if (bp->b_flags&B_ERROR) {
		brelse(bp);
		iput(ip);
		return (NULL);
	}
	dp = &bp->b_un.b_dino[itoo(ino)];
	ip->i_mode  = dp->di_mode;
	ip->i_nlink = dp->di_nlink;
	ip->i_size  = dp->di_size;
	iexpand(ip->i_addr, dp->di_addr);

	brelse(bp);

	return (ip);
}

/*
 * Decrement reference count of
 * an inode structure.
 * On the last reference,
 * write the inode out and if necessary,
 * truncate and deallocate the file.
 */
iput(ip)
register struct inode *ip;
{

	if (ip->i_count == 1) {
		ip->i_flag |= ILOCK;
		if (ip->i_nlink == 0) {
			itrunc(ip);
			ip->i_mode = 0;
			ip->i_flag |= IUPD|ICHG;
			ifree(ip->i_mp, ip->i_number);
		}
		iupdat(ip);
		ip->i_flag = 0;
		ip->i_number = 0;
		ip->i_mp = NULL;
	}
	ip->i_count--;
	prele(ip);
}

/*
 * Check accessed and update flags on an inode structure.
 * If any is on, update the inode
 */
iupdat(ip)
	struct inode *ip;
{
	struct buf *bp;
	struct dinode *dp;

	if (ip->i_flag & (IUPD|ICHG)) {
		if (ip->i_mp->m_filsys.s_ronly)
			return;
		bp = bread(ip->i_mp->m_dev, itod(ip->i_number));
		if (bp->b_flags & B_ERROR) {
			brelse(bp);
			return;
		}
		dp = &bp->b_un.b_dino[itoo(ip->i_number)];
		dp->di_mode  = ip->i_mode;
		dp->di_nlink = ip->i_nlink;
		dp->di_size  = ip->i_size;
		icompress(dp->di_addr, ip->i_addr);

		ip->i_flag &= ~(IUPD|ICHG);
		bdwrite(bp);
	}
}

/*
 * Free all the disk blocks associated
 * with the specified inode structure.
 * The blocks of the file are removed
 * in reverse order. This FILO
 * algorithm will tend to maintain
 * a contiguous free list much longer
 * than FIFO.
 */
itrunc(ip)
	struct inode *ip;
{
	daddr_t bn;
	struct mount *mp = ip->i_mp;
	int i = NADDR-1;

	if (ip->i_mode & (IFCHR&IFBLK)) /* XXX chr or blk ? */
		return;

	do {
		bn = ip->i_addr[i];
		if (bn) {
			ip->i_addr[i] = 0;
			switch (i) {
			case NADDR-3:
				tloop(mp, bn, 0, 0);
				break;

			case NADDR-2:
				tloop(mp, bn, 1, 0);
				break;

			case NADDR-1:
				tloop(mp, bn, 1, 1);
				break;

			default:
				free(mp, bn);
			}
		}
	} while (--i >= 0);

	ip->i_size = 0;
	ip->i_flag |= ICHG|IUPD;
}

tloop(mp, bn, f1, f2)
	struct mount *mp;
	daddr_t bn;
{
	daddr_t *bap, nb;
	int i = NINDIR-1;
	struct buf *bp = NULL;

	do {
		if (bp == NULL) {
			bp = bread(mp->m_dev, bn);
			if (bp->b_flags & B_ERROR) {
				brelse(bp);
				return;
			}
			bap = bp->b_un.b_daddr;
		}
		nb = bap[i];
		if (nb) {
			if (f1) {
				brelse(bp);
				tloop(mp, nb, f2, 0);
				bp = NULL;
			} else
				free(mp, nb);
		}
	} while (--i >= 0);

	if (bp)
		brelse(bp);

	free(mp, bn);
}

/*
 * Make a new file.
 */
struct inode *
maknode(mode)
{
	register struct inode *ip;

	ip = ialloc(u.u_pdir->i_mp);
	if (ip == NULL) {
		iput(u.u_pdir);
		return (NULL);
	}
	ip->i_flag |= IUPD|ICHG;
	ip->i_mode = mode;
	ip->i_nlink = 1;
	wdir(ip->i_number);

	return (ip);
}

/*
 * Write a directory entry with
 * parameters left as side effects
 * to a call to namei.
 */
wdir(inum)
{
	u.u_dent.d_ino = inum;
	bcopy(u.u_dbuf, u.u_dent.d_name, DIRSIZ);

	u.u_count = sizeof(struct direct);
	u.u_base = &u.u_dent;
	u.u_segflg = 0;
	writei(u.u_pdir, 0);

	iput(u.u_pdir);
}

icompress(pe, pi)
	register char *pe, *pi;
{
	int n = NADDR;

	do {
		if (pi[3])
			goto oops;
		pe[0] = pi[0];
		pe[1] = pi[1];
		pe[2] = pi[2];
		pi += 4;
		pe += 3;
	} while (--n);

	return;

oops:
	panic("icompress");
}

iexpand(pi, pe)
	register char *pi, *pe;
{
	int n = NADDR;

	do {
		pi[0] = pe[0];
		pi[1] = pe[1];
		pi[2] = pe[2];
		pi[3] = 0;
		pi += 4;
		pe += 3;
	} while (--n);
}

